import {
  createReadableStream,
  createWritableStream,
  timeout
} from "@datastream/core";
const defaults = {
  rateLimit: 0.01,
  rateLimitTimestamp: 0,
  dataPath: void 0,
  nextPath: void 0,
  qs: void 0,
  method: "GET",
  headers: {
    Accept: "application/json",
    "Accept-Encoding": "br, gzip, deflate"
  }
};
const mergeOptions = (options = {}) => {
  return {
    ...defaults,
    ...options,
    headers: { ...defaults.headers, ...options.headers }
  };
};
const fetchSetDefaults = (options) => {
  Object.assign(defaults, mergeOptions(options));
};
const fetchWritableStream = async (options, streamOptions = {}) => {
  const body = createReadableStream();
  const value = await fetchRateLimit({
    ...options,
    body,
    duplex: "half",
    signal: streamOptions.signal
  });
  const write = (chunk) => {
    body.push(chunk);
  };
  const stream = createWritableStream(write, streamOptions);
  stream.result = () => ({ key: "output", value });
  return stream;
};
const fetchRequestStream = fetchWritableStream;
const fetchReadableStream = (fetchOptions, streamOptions) => {
  if (!Array.isArray(fetchOptions))
    fetchOptions = [fetchOptions];
  return createReadableStream(
    fetchGenerator(fetchOptions, streamOptions),
    streamOptions
  );
};
const fetchResponseStream = fetchReadableStream;
async function* fetchGenerator(fetchOptionsArray, streamOptions) {
  const requests = [];
  for (let options of fetchOptionsArray) {
    options = mergeOptions(options);
    if (options.qs) {
      options.url += ("?" + new URLSearchParams(options.qs)).replaceAll(
        "+",
        "%20"
      );
    }
    requests.push(fetchUnknown(options, streamOptions));
  }
  for (const request of requests) {
    const response = await request;
    for await (const chunk of response) {
      yield chunk;
    }
  }
}
const jsonContentTypeRegExp = /^application\/(.+\+)?json($|;.+)/;
const fetchUnknown = async (options, streamOptions) => {
  const response = await fetchRateLimit(options, streamOptions);
  if (jsonContentTypeRegExp.test(response.headers.get("Content-Type"))) {
    options.prefetchResponse = response;
    return fetchJson(options, streamOptions);
  }
  return response.body;
};
const nextLinkRegExp = /<(.*?)>; rel="next"/;
async function* fetchJson(options, streamOptions) {
  const { dataPath, nextPath } = options;
  while (options.url) {
    const response = options.prefetchResponse ?? await fetchRateLimit(options, streamOptions);
    delete options.prefetchResponse;
    const body = await response.json();
    options.url = nextPath && pickPath(body, nextPath);
    options.url ??= parseLink(response.headers);
    const data = pickPath(body, dataPath);
    if (Array.isArray(data)) {
      for (const item of data) {
        yield item;
      }
    } else {
      yield data;
    }
  }
}
const parseLink = (headers) => {
  const link = headers.get("Link");
  if (link) {
    return link.match(nextLinkRegExp)?.[1];
  }
};
const fetchRateLimit = async (options, streamOptions = {}) => {
  options.rateLimitTimestamp ??= 0;
  const now = Date.now();
  if (now < options.rateLimitTimestamp) {
    await timeout(options.rateLimitTimestamp - now, streamOptions);
  }
  options.rateLimitTimestamp = now + 1e3 * options.rateLimit;
  options = mergeOptions(options);
  const response = await fetch(options.url, {
    ...options,
    signal: streamOptions.signal
  });
  if (!response.ok) {
    if (response.statusCode === 429) {
      return fetchRateLimit(options, streamOptions);
    }
    throw new Error("fetch", {
      cause: { request: options, response }
    });
  }
  return response;
};
const pickPath = (obj, path = "") => {
  if (path === "")
    return obj;
  if (!Array.isArray(path))
    path = path.split(".");
  return path.slice(0).reduce((a, b) => {
    return a[b];
  }, obj);
};
var fetch_default = {
  setDefaults: fetchSetDefaults,
  readableStream: fetchReadableStream,
  responseStream: fetchReadableStream
};
export {
  fetch_default as default,
  fetchRateLimit,
  fetchReadableStream,
  fetchRequestStream,
  fetchResponseStream,
  fetchSetDefaults,
  fetchWritableStream
};
